from pyske.core.support.errors import IllFormedError, NotEqualSizeError, ApplicationError, EmptyError, NotSameTagError
from pyske.core.tree.ltree import Segment, TaggedValue
import pytest

def test_has_critical_empty():
    seg = Segment()
    exp = False
    res = seg.has_critical()
    assert exp == res


def test_has_critical_no():
    val1 = TaggedValue(1,"N")
    val2 = TaggedValue(2,"L")
    seg = Segment([val1, val2])
    exp = False
    res = seg.has_critical()
    assert exp == res


def test_has_critical_yes():
    val1 = TaggedValue(1,"N")
    val2 = TaggedValue(2,"L")
    val3 = TaggedValue(3,"C")
    seg = Segment([val1, val2, val3])
    exp = True
    res = seg.has_critical()
    assert exp == res


	
# -------------------------- #

def test_map_local_empty():
    seg = Segment()
    res = seg.map_local(lambda x : x + 1, lambda x : x - 1 )
    exp = Segment()
    assert exp == res


def test_map_local():
    seg = Segment([TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"C")])
    res = seg.map_local(lambda x : x + 1, lambda x : x - 1 )
    exp = Segment([TaggedValue(0,"N"), TaggedValue(3,"L"), TaggedValue(2,"C")])
    assert exp == res


# -------------------------- #

def test_reduce_local_empty():
    sum3 = lambda x,y,z : x + y + z
    id_f = lambda x : x
    seg = Segment()
    with pytest.raises(AssertionError):
        seg.reduce_local(sum3, id_f, sum3, sum3)


def test_reduce_local_illformed():
    sum3 = lambda x,y,z : x + y + z
    id_f = lambda x : x
    seg = Segment([TaggedValue(1,"N"), TaggedValue(2,"C")])
    with pytest.raises(IllFormedError):
        seg.reduce_local(sum3, id_f, sum3, sum3)


def test_reduce_local_node():
    sum3 = lambda x,y,z : x + y + z
    id_f = lambda x : x
    seg = Segment([TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"C")])
    res = seg.reduce_local(sum3, id_f, sum3, sum3)
    exp = TaggedValue(6,"N")
    assert res == exp


def test_reduce_local_leaf():
    sum3 = lambda x,y,z : x + y + z
    id_f = lambda x : x
    seg = Segment([TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"L")])
    res = seg.reduce_local(sum3, id_f, sum3, sum3)
    exp = TaggedValue(6,"L")
    assert res == exp


# -------------------------- #

def test_reduce_global_has_critical():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(1,"N"), TaggedValue(2, "L"), TaggedValue(2,"C")])
    with pytest.raises(AssertionError):
        seg.reduce_global(sum3)


def test_reduce_global_empty():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment()
    with pytest.raises(AssertionError):
        seg.reduce_global(sum3)


def test_reduce_global_illformed():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(1,"N"), TaggedValue(2, "L")])
    with pytest.raises(IllFormedError):
        seg.reduce_global(sum3)


def test_reduce_global_leaf():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(2, "L")])
    res = seg.reduce_global(sum3)
    exp = 2
    assert res == exp


def test_reduce_global_node():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(2, "N"),TaggedValue(2, "L"),TaggedValue(2, "L")])
    res = seg.reduce_global(sum3)
    exp = 6
    assert res == exp


# -------------------------- #

def test_uacc_local_empty():
    sum3 = lambda x,y,z : x + y + z
    id_f = lambda x : x
    seg = Segment()
    with pytest.raises(AssertionError):
        seg.uacc_local(sum3, id_f, sum3, sum3)


def test_uacc_local_illformed():
    sum3 = lambda x,y,z : x + y + z
    id_f = lambda x : x
    seg = Segment([TaggedValue(1,"N"), TaggedValue(2,"C")])
    with pytest.raises(IllFormedError):
        seg.uacc_local(sum3, id_f, sum3, sum3)


def test_uacc_local_node():
    sum3 = lambda x,y,z : x + y + z
    id_f = lambda x : x
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"C")])
    res = seg.uacc_local(sum3, id_f, sum3, sum3)
    exp = (TaggedValue(9,"N"), Segment([None,  TaggedValue(2,"L"), None, TaggedValue(2,"L"), None]))
    assert res == exp

def test_uacc_local_leaf():
    sum3 = lambda x,y,z : x + y + z
    id_f = lambda x : x
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"L")])
    res = seg.uacc_local(sum3, id_f, sum3, sum3)
    exp = (TaggedValue(9,"L"), Segment([TaggedValue(9,"N"),  TaggedValue(2,"L"), TaggedValue(6,"N"), TaggedValue(2,"L"), TaggedValue(3,"L")]))
    assert res == exp

# -------------------------- #
phi = lambda b : (1, 0, 0, 1)

def k(l, b, r):
    (ll, ls) = l
    (rl, rs) = r
    return (ls, ls + 1 + rs)


def psi_l(l, b, r):
    (l0, l1, l2, l3) = l
    (b0, b1, b2, b3) = b
    (rl, rs) = r
    res_0 = 0
    res_1 = b0 + b1
    res_2 = (b0 + b1) * l3 + b1 * (1 + rs) + b2
    res_3 = l3 + 1 + rs + b3
    return (res_0, res_1, res_2, res_3)


def psi_r(l, b, r):
    (ll, ls) = l
    (b0, b1, b2, b3) = b
    (r0, r1, r2, r3) = r
    res_0 = 0
    res_1 = b1
    res_2 = b1 * r3 + b0 * ls + b1 * (1 + ls) + b2
    res_3 = r3 + 1 + ls + b3
    return (res_0, res_1, res_2, res_3)


def psi_n(l, b, r):
    (ll, ls) = l
    (b0, b1, b2, b3) = b
    (rl, rs) = r
    res_1 = b0 * ls + b1 * (ls + rs + 1) + b2
    res_2 = ls + 1 + rs + b3
    return (res_1, res_2)


def test_uacc_local_prefix_1():
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"C"), TaggedValue((0,1),"L")])
    res = seg.uacc_local(k, phi, psi_l, psi_r)
    exp = (TaggedValue((0,1,1,4),"N"), Segment([None, None, TaggedValue((0,1),"L")]))
    assert res == exp


def test_uacc_local_prefix_2():
    seg = Segment([TaggedValue(4,"N"),  TaggedValue((0,1),"L"), TaggedValue((0,1),"L")])
    res = seg.uacc_local(k, phi, psi_l, psi_r)
    exp = (TaggedValue((1,3),"L"), Segment([TaggedValue((1,3),"N"), TaggedValue((0,1),"L"), TaggedValue((0,1),"L")]))
    assert res == exp


def test_uacc_local_prefix_3():
    seg = Segment([TaggedValue(5,"N"),  TaggedValue((0,1),"L"), TaggedValue((0,1),"L")])
    res = seg.uacc_local(k, phi, psi_l, psi_r)
    exp = (TaggedValue((1,3),"L"), Segment([TaggedValue((1,3),"N"), TaggedValue((0,1),"L"), TaggedValue((0,1),"L")]))
    assert res == exp


def test_uacc_global_prefix():
    gt = Segment([TaggedValue((0,1,1,4),"N"), TaggedValue((1,3),"L"),TaggedValue((1,3),"L")])
    res = gt.uacc_global(psi_n)
    exp = Segment([TaggedValue((8,11),"N"), TaggedValue((1,3),"L"),TaggedValue((1,3),"L")])
    assert res == exp


def test_uacc_update_prefix():
    seg = Segment([TaggedValue(1,"N"),TaggedValue(2,"C"),TaggedValue((0,1),"L")])
    seg2 = Segment([None, None ,TaggedValue((0,1),"L")])
    lc = (1,3)
    rc = (1,3)
    res = seg.uacc_update(seg2, k, lc, rc)
    exp = Segment([TaggedValue((5,9),"N"),TaggedValue((1,5),"C"),TaggedValue((0,1),"L")])
    assert res == exp


# -------------------------- #

def test_uacc_global_has_critical():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(1,"N"), TaggedValue(2,"C"), TaggedValue(2,"L")])
    with pytest.raises(AssertionError):
        seg.uacc_global(sum3)


def test_uacc_global_empty():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment()
    exp = Segment()
    res = seg.uacc_global(sum3)
    assert exp == res


def test_uacc_global_illformed():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(1,"N"), TaggedValue(2,"L")])
    with pytest.raises(IllFormedError):
        seg.uacc_global(sum3)


def test_uacc_global_leaf():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(3,"L")])
    res = seg.uacc_global(sum3)
    exp = Segment([TaggedValue(3,"L")])
    assert res == exp


def test_uacc_global_node():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"L")])
    res = seg.uacc_global(sum3)
    exp = Segment([TaggedValue(9,"N"),  TaggedValue(2,"L"), TaggedValue(6,"N"), TaggedValue(2,"L"), TaggedValue(3,"L")])
    assert res == exp

# -------------------------- #

def test_uacc_update_empty():
    seg = Segment()
    sum3 = lambda x,y,z : x + y + z
    gt = Segment()
    lc = 1
    rc = 2
    res = seg.uacc_update(gt, sum3, lc, rc)
    exp = Segment()
    assert res == exp


def test_uacc_update_not_same_size():
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"L")])
    sum3 = lambda x,y,z : x + y + z
    gt = Segment()
    lc = 1
    rc = 2
    with pytest.raises(AssertionError):
        seg.uacc_update(gt, sum3, lc, rc)


def test_uacc_update_illformed_node():
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"N"), TaggedValue(2,"N")])
    sum3 = lambda x,y,z : x + y + z
    gt = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(2,"L")])
    lc = 1
    rc = 2
    with pytest.raises(IllFormedError):
        seg.uacc_update(gt, sum3, lc, rc)


def test_uacc_update_illformed_critical():
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"C"),TaggedValue(1,"N")])
    sum3 = lambda x,y,z : x + y + z
    gt = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(1,"L")])
    lc = 1
    rc = 2
    with pytest.raises(IllFormedError):
        seg.uacc_update(gt, sum3, lc, rc)


def test_uacc_update_leaf():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"L")])
    gt =  Segment([TaggedValue(9,"N"),  TaggedValue(2,"L"), TaggedValue(6,"N"), TaggedValue(2,"L"), TaggedValue(3,"L")])
    lc = 1
    rc = 2
    res = seg.uacc_update(gt, sum3, lc, rc)
    exp = Segment([TaggedValue(9,"N"),  TaggedValue(2,"L"), TaggedValue(6,"N"), TaggedValue(2,"L"), TaggedValue(3,"L")])
    assert res == exp


def test_uacc_update_node():
    sum3 = lambda x,y,z : x + y + z
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"C")])
    gt = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"C")])
    lc = 1
    rc = 2
    res = seg.uacc_update(gt, sum3, lc, rc)
    exp = Segment([TaggedValue(12,"N"),  TaggedValue(2,"L"), TaggedValue(9,"N"), TaggedValue(2,"L"), TaggedValue(6,"C")])
    assert res == exp

# -------------------------- #

def test_dacc_path_empty():
    seg = Segment()
    sum2 = lambda x,y : x + y
    id_f = lambda x : x
    with pytest.raises(AssertionError):
        seg.dacc_path(id_f, id_f, sum2)


def test_dacc_path_has_no_critical():
    seg = Segment([TaggedValue(1,"N"),TaggedValue(2,"L"),TaggedValue(3,"L")])
    sum2 = lambda x,y : x + y
    id_f = lambda x : x
    with pytest.raises(ApplicationError):
        seg.dacc_path(id_f, id_f, sum2)


def test_dacc_path_gt_node():
    seg = Segment([TaggedValue(1,"N"),  TaggedValue(2,"L"), TaggedValue(1,"N"), TaggedValue(2,"L"), TaggedValue(3,"C")])
    sum2 = lambda x,y : x + y
    id_f = lambda x : x

    res = seg.dacc_path(id_f, id_f, sum2)
    exp = TaggedValue((5,5),"N")
    assert res == exp

# -------------------------- #

def test_dacc_global_double_leaf():
    seg = Segment([TaggedValue(1,"L"), TaggedValue(1,"L")])
    sum2 = lambda x,y : x + y
    c = 2
    with pytest.raises(IllFormedError):
        seg.dacc_global(sum2, c)


def test_dacc_global_empty():
    seg = Segment()
    sum2 = lambda x,y : x + y
    c = 2
    res = seg.dacc_global(sum2, c)
    exp = Segment()
    assert res == exp


def test_dacc_global_has_critical():
    seg = Segment([TaggedValue(1,"N"), TaggedValue(1,"C"), TaggedValue(1,"L")])
    sum2 = lambda x,y : x + y
    c = 2
    with pytest.raises(AssertionError):
        seg.dacc_global(sum2, c)


def test_dacc_global():
    sum2 = lambda x,y : x + y
    c = 2
    seg = Segment([TaggedValue((1,1),"N"),  TaggedValue((2,3),"L"), TaggedValue((1,2),"N"), TaggedValue((2,2),"L"), TaggedValue((3,1),"L")])
    res = seg.dacc_global(sum2, c)
    exp = Segment([TaggedValue(2,"N"),  TaggedValue(3,"L"), TaggedValue(3,"N"), TaggedValue(4,"L"), TaggedValue(5,"L")])
    assert res == exp

# -------------------------- #

def test_dacc_local_empty():
    seg = Segment()
    sum2 = lambda x, y: x + y
    c = 4
    res = seg.dacc_local(sum2, sum2, c)
    exp = Segment()
    assert res == exp


def test_dacc_local_stack_empty_leaf():
    seg = Segment([TaggedValue(2,"L"),TaggedValue(2,"L")])
    sum2 = lambda x, y: x + y
    c = 4
    with pytest.raises(IllFormedError):
        seg.dacc_local(sum2, sum2, c)


def test_dacc_local_stack_empty_critical():
    seg = Segment([TaggedValue(2,"C"),TaggedValue(2,"L")])
    sum2 = lambda x, y: x + y
    c = 4
    with pytest.raises(IllFormedError):
        seg.dacc_local(sum2, sum2, c)


def test_dacc_local():
    seg = Segment([TaggedValue(231,"N"),  TaggedValue(169,"L"), TaggedValue(478,"N"), TaggedValue(634,"L"), TaggedValue(2,"C")])
    sum2 = lambda x, y: x + y
    min2 = lambda x, y: x - y
    c = 400
    res = seg.dacc_local(sum2, min2, c)
    exp = Segment([TaggedValue(400,"N"),  TaggedValue(631,"L"), TaggedValue(169,"N"), TaggedValue(647,"L"), TaggedValue(-309,"C")])
    assert res == exp

# -------------------------- #

def test_get_left_has_critical():
    gt = Segment([TaggedValue(2,"N"),TaggedValue(1,"C"),TaggedValue(2,"L")])
    i = 0
    with pytest.raises(AssertionError):
        gt.get_left(i)


def test_get_left_is_leaf():
    gt = Segment([TaggedValue(2,"N"),TaggedValue(1,"L"),TaggedValue(2,"L")])
    i = 1
    with pytest.raises(AssertionError):
        gt.get_left(i)


def test_get_left_illformed():
    gt = Segment([TaggedValue(2,"N")])
    i = 0
    with pytest.raises(AssertionError):
        gt.get_left(i)


def test_get_left():
    gt = Segment([TaggedValue(2,"N"),TaggedValue(1,"L"),TaggedValue(2,"L")])
    i = 0
    res = gt.get_left(i)
    exp = TaggedValue(1,"L")
    assert res == exp


# -------------------------- #

def test_get_right_has_critical():
    gt = Segment([TaggedValue(2,"N"),TaggedValue(1,"C"),TaggedValue(2,"L")])
    i = 0
    with pytest.raises(AssertionError):
        gt.get_right(i)


def test_get_right_is_leaf():
    gt = Segment([TaggedValue(2,"N"),TaggedValue(1,"L"),TaggedValue(2,"L")])
    i = 1
    with pytest.raises(AssertionError):
        gt.get_right(i)


def test_get_right_illformed():
    gt = Segment([TaggedValue(2,"N"),TaggedValue(1,"L")])
    i = 0
    with pytest.raises(AssertionError):
        gt.get_right(i)


def test_get_right_direct():
    gt = Segment([TaggedValue(2,"N"),TaggedValue(1,"L"),TaggedValue(2,"L")])
    i = 0
    res = gt.get_right(i)
    exp = TaggedValue(2,"L")
    assert res == exp


def test_get_right_not_direct():
    gt = Segment([TaggedValue(1,"N"),TaggedValue(3,"N"),TaggedValue(8,"L"),TaggedValue(4,"L"),TaggedValue(2,"L")])
    i = 0
    res = gt.get_right(i)
    exp = TaggedValue(2,"L")
    assert res == exp


# -------------------------- #

def test_zip_not_equal_size_error():
    seg1 = Segment([TaggedValue(2,"N"),TaggedValue(1,"L"),TaggedValue(2,"L")])
    seg2 = Segment([TaggedValue(2,"L")])
    with pytest.raises(AssertionError):
        seg1.zip(seg2)


def test_zip_not_same_tag_error():
    seg1 = Segment([TaggedValue(2,"N"),TaggedValue(1,"L"),TaggedValue(2,"L")])
    seg2 = Segment([TaggedValue(2,"N"),TaggedValue(1,"C"),TaggedValue(2,"L")])
    with pytest.raises(NotSameTagError):
        seg1.zip(seg2)


def test_zip():
    seg1 = Segment([TaggedValue(1,"N"),TaggedValue(2,"L"),TaggedValue(3,"L")])
    seg2 = Segment([TaggedValue(4,"N"),TaggedValue(5,"L"),TaggedValue(6,"L")])
    res = seg1.zip(seg2)
    exp = Segment([TaggedValue((1,4),"N"),TaggedValue((2,5),"L"),TaggedValue((3,6),"L")])
    assert res == exp

# -------------------------- #


def test_zipwith_not_equal_size_error():
    sum2 = lambda x,y : x + y
    seg1 = Segment([TaggedValue(2,"N"),TaggedValue(1,"L"),TaggedValue(2,"L")])
    seg2 = Segment([TaggedValue(2,"L")])
    with pytest.raises(AssertionError):
        seg1.map2(sum2, seg2)


def test_zipwith_not_same_tag_error():
    sum2 = lambda x,y : x + y
    seg1 = Segment([TaggedValue(2,"N"),TaggedValue(1,"L"),TaggedValue(2,"L")])
    seg2 = Segment([TaggedValue(2,"N"),TaggedValue(1,"C"),TaggedValue(2,"L")])
    with pytest.raises(NotSameTagError):
        seg1.map2(sum2, seg2)


def test_zipwith():
    sum2 = lambda x,y : x + y
    seg1 = Segment([TaggedValue(1,"N"),TaggedValue(2,"L"),TaggedValue(3,"L")])
    seg2 = Segment([TaggedValue(4,"N"),TaggedValue(5,"L"),TaggedValue(6,"L")])
    res = seg1.map2(sum2, seg2)
    exp = Segment([TaggedValue(5,"N"),TaggedValue(7,"L"),TaggedValue(9,"L")])
    assert res == exp

# -------------------------- #

def test_from_str():
    s = "[(11^N); (2^L); (3^C)]"
    res = Segment.from_str(s)
    exp = Segment([TaggedValue(11,"N"),TaggedValue(2,"L"),TaggedValue(3,"C")])
    assert res == exp

# -------------------------- #


from pyske.core.list.slist import SList
from pyske.core.support.parallel import *


class PList:
    """Distributed lists"""

    def __init__(self):
        self.__content = SList([])
        self.__global_size = 0
        self.__local_size = 0
        self.__start_index = 0
        self.__distribution = [0 for i in range(0, nprocs)]

    def length(self):
        return self.__global_size

    @staticmethod
    def init(f, size):
        assert (size >= 0)
        p = PList()
        p.__global_size = size
        p.__local_size = local_size(size)
        p.__distribution = [local_size_pid(i, size) for i in range(0, nprocs)]
        p.__start_index = SList(p.__distribution).scanl(lambda x, y: x + y, 0)[pid]
        p.__content = SList([f(i) for i in range(p.__start_index, p.__start_index + p.__local_size)])
        p.__distribution = [local_size_pid(i, size) for i in range(0, nprocs)]
        return p

    def map(self, f):
        return PList.init(lambda i: f(self.__content[i - self.__start_index]), self.__global_size)

    def mapi(self, f):
        return PList.init(lambda i: f(i, self.__content[i - self.__start_index]), self.__global_size)

    def map2(self, f, pl):
        assert (self.__distribution == pl.__distribution)
        return PList.init(lambda i: f(self.__content[i - self.__start_index], pl.__content[i - self.__start_index]),
                          self.__global_size)

    def zip(self, pl):
        return self.map2(lambda x, y: (x, y), pl)

    def get_partition(self):
        p = PList()
        p.__content = [self.__content]
        p.__global_size = nprocs
        p.__local_size = 1
        p.__start_index = pid
        p.__distribution = [1 for i in range(0, nprocs)]
        return p

    def flatten(self):
        p = PList()
        p.__content = self.__content.reduce(lambda x, y: x + y, [])
        p.__local_size = len(p.__content)
        p.__distribution = comm.allgather(p.__local_size)
        p.__start_index = SList(p.__distribution).scanl(lambda x, y: x + y, 0)[pid]
        p.__global_size = SList(p.__distribution).reduce(lambda x, y: x + y)
        return p

    def reduce(self, op, e=None):
        if e is None:
            assert (self.__global_size >= 1)
            partial = None if self.__local_size == 0 else SList(self.__content).reduce(op)
            partials = SList(comm.allgather(partial)).filter(lambda x: x is not None)
        else:
            # assert: (op, e) form a monoid
            partial = SList(self.__content).reduce(op, e)
            partials = SList(comm.allgather(partial))
        return partials.reduce(op, e)

    def __get_shape(self):
        p = PList()
        p.__local_size = self.__local_size
        p.__global_size = self.__global_size
        p.__distribution = self.__distribution
        p.__start_index = self.__start_index
        return p

    def scanr(self, op):
        assert (self.__global_size > 0)
        p = self.__get_shape()
        partials = self.__content.scanr(op)
        last = partials[self.__local_size - 1]
        acc, _ = scan(op, last)
        if pid != 0:
            for i in range(0, len(partials)):
                partials[i] = op(acc, partials[i])
        p.__content = partials
        return p

    def scanl(self, op, e):
        p = self.__get_shape()
        partials, last = self.__content.scanl_last(op, e)
        acc, _ = scan(op, last)
        if pid != 0:
            for i in range(0, len(partials)):
                partials[i] = op(acc, partials[i])
        p.__content = partials
        return p

    def scanl_last(self, op, e):
        p = self.__get_shape()
        partials, last = self.__content.scanl_last(op, e)
        acc, red = scan(op, last)
        if pid != 0:
            for i in range(0, len(partials)):
                partials[i] = op(acc, partials[i])
        p.__content = partials
        return p, red

    @staticmethod
    def from_seq(l):
        p = PList()
        if pid == 0:
            p.__content = SList(l)
            p.__distribution = [len(l)] + [0 for i in range(1, nprocs)]
        else:
            p.__content = []
        p.__distribution = comm.bcast(p.__distribution, 0)
        p.__local_size = p.__distribution[pid]
        p.__global_size = p.__distribution[0]
        p.__start_index = 0
        return p

    def to_seq(self):
        return self.get_partition().reduce(lambda x, y: x + y, [])

    def __str__(self):
        return "pid[" + str(pid) + "]:\n" + \
               "  global_size: " + str(self.__global_size) + "\n" + \
               "  local_size: " + str(self.__local_size) + "\n" + \
               "  start_index: " + str(self.__start_index) + "\n" + \
               "  distribution: " + str(self.__distribution) + "\n" + \
               "  content: " + str(self.__content) + "\n"



import pytest

from pyske.core.tree.ptree import *
from pyske.core.support.parallel import *
from pyske.core.tree.ltree import *

def illformed_ltree():
    seg1 = Segment([TaggedValue(13, "C")])
    seg3 = Segment([TaggedValue(72, "N"), TaggedValue(92, "L"), TaggedValue(42, "L")])
    lt = LTree([seg1, seg3])
    with pytest.raises(IllFormedError):
        PTree(lt)


def test_to_seq():
    seg1 = Segment([TaggedValue(13, "C")])
    seg2 = Segment([TaggedValue(31, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    seg3 = Segment([TaggedValue(72, "N"), TaggedValue(92, "L"), TaggedValue(42, "L")])
    lt = LTree([seg1, seg2, seg3])
    pt = PTree(lt)
    res = pt.to_seq()
    exp = lt if pid == 0 else None
    assert res == exp


def test_map_empty():
    pt = PTree()
    id_f = lambda x: x
    exp = PTree()
    res = pt.map(id_f, id_f)
    assert exp == res


def test_map_not_empty():
    seg1 = Segment([TaggedValue(13, "C")])
    seg2 = Segment([TaggedValue(31, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    seg3 = Segment([TaggedValue(72, "N"), TaggedValue(92, "L"), TaggedValue(42, "L")])
    lt = LTree([seg1, seg2, seg3])
    pt = PTree(lt)

    res = pt.map(lambda x: x + 1, lambda x: x - 1).to_seq()

    seg1_exp = Segment([TaggedValue(12, "C")])
    seg2_exp = Segment([TaggedValue(30, "N"), TaggedValue(48, "L"), TaggedValue(33, "L")])
    seg3_exp = Segment([TaggedValue(71, "N"), TaggedValue(93, "L"), TaggedValue(43, "L")])
    exp = LTree([seg1_exp, seg2_exp, seg3_exp]) if pid == 0 else None

    assert res == exp


# -------------------------- #

def test_reduce_empty():
    pt = PTree()
    id_f = lambda x: x
    sum3 = lambda x, y, z: x + y + z
    with pytest.raises(AssertionError):
        pt.reduce(sum3, id_f, sum3, sum3, sum3)


def test_reduce():
    seg1 = Segment([TaggedValue(13, "C")])
    seg2 = Segment([TaggedValue(31, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    seg3 = Segment([TaggedValue(72, "N"), TaggedValue(92, "L"), TaggedValue(42, "L")])
    lt = LTree([seg1, seg2, seg3])
    pt = PTree(lt)
    id_f = lambda x: x
    sum3 = lambda x, y, z: x + y + z
    res = pt.reduce(sum3, id_f, sum3, sum3, sum3)
    exp = lt.reduce(sum3, id_f, sum3, sum3, sum3) if pid == 0 else None
    assert res == exp


# -------------------------- #

def test_uacc_empty():
    pt = PTree()
    id_f = lambda x: x
    sum3 = lambda x, y, z: x + y + z
    with pytest.raises(AssertionError):
        pt.uacc(sum3, id_f, sum3, sum3, sum3)


def test_uacc():
    seg1 = Segment([TaggedValue(13, "C")])
    seg2 = Segment([TaggedValue(31, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    seg3 = Segment([TaggedValue(72, "N"), TaggedValue(92, "L"), TaggedValue(42, "L")])
    lt = LTree([seg1, seg2, seg3])
    pt = PTree(lt)
    id_f = lambda x: x
    sum3 = lambda x, y, z: x + y + z
    res = pt.uacc(sum3, id_f, sum3, sum3, sum3).to_seq()

    seg1_exp = Segment([TaggedValue(13 + 31 + 47 + 32 + 72 + 92 + 42, "C")])
    seg2_exp = Segment([TaggedValue(31 + 47 + 32, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    seg3_exp = Segment([TaggedValue(72 + 92 + 42, "N"), TaggedValue(92, "L"), TaggedValue(42, "L")])
    exp = LTree([seg1_exp, seg2_exp, seg3_exp]) if pid == 0 else None
    assert exp == res

# -------------------------- #

def test_dacc():
    sum2 = lambda x, y: x + y
    c = 0
    id_f = lambda x: x
    seg1 = Segment([TaggedValue(13, "C")])
    seg2 = Segment([TaggedValue(31, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    seg3 = Segment([TaggedValue(72, "N"), TaggedValue(92, "L"), TaggedValue(42, "L")])
    lt = LTree([seg1, seg2, seg3])
    res = PTree(lt).dacc(sum2, sum2, c, id_f, id_f, sum2, sum2).to_seq()

    seg1_exp = Segment([TaggedValue(0, "C")])
    seg2_exp = Segment([TaggedValue(13, "N"), TaggedValue(13 + 31, "L"), TaggedValue(13 + 31, "L")])
    seg3_exp = Segment([TaggedValue(13, "N"), TaggedValue(13 + 72, "L"), TaggedValue(13 + 72, "L")])
    exp = LTree([seg1_exp, seg2_exp, seg3_exp]) if pid == 0 else None

    assert res == exp


# -------------------------- #

def test_zip_not_same_size():
    seg11 = Segment([TaggedValue(13, "C")])
    seg21 = Segment([TaggedValue(31, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    seg31 = Segment([TaggedValue(72, "N"), TaggedValue(92, "L"), TaggedValue(42, "L")])
    lt1 = LTree([seg11, seg21, seg31])
    seg22 = Segment([TaggedValue(31, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    lt2 = LTree([seg22])
    pt1 = PTree(lt1)
    pt2 = PTree(lt2)
    with pytest.raises(AssertionError):
        pt1.zip(pt2)


def test_zip():
    seg11 = Segment([TaggedValue(1, "C")])
    seg21 = Segment([TaggedValue(1, "N"), TaggedValue(1, "L"), TaggedValue(1, "L")])
    seg31 = Segment([TaggedValue(1, "N"), TaggedValue(1, "L"), TaggedValue(1, "L")])
    lt1 = LTree([seg11, seg21, seg31])
    seg12 = Segment([TaggedValue(2, "C")])
    seg22 = Segment([TaggedValue(2, "N"), TaggedValue(2, "L"), TaggedValue(2, "L")])
    seg32 = Segment([TaggedValue(2, "N"), TaggedValue(2, "L"), TaggedValue(2, "L")])
    lt2 = LTree([seg12, seg22, seg32])
    pt1 = PTree(lt1)
    pt2 = PTree(lt2)
    res = pt1.zip(pt2).to_seq()
    seg1_exp = Segment([TaggedValue((1, 2), "C")])
    seg2_exp = Segment([TaggedValue((1, 2), "N"), TaggedValue((1, 2), "L"), TaggedValue((1, 2), "L")])
    seg3_exp = Segment([TaggedValue((1, 2), "N"), TaggedValue((1, 2), "L"), TaggedValue((1, 2), "L")])
    exp = LTree([seg1_exp, seg2_exp, seg3_exp]) if pid == 0 else None
    assert res == exp


# -------------------------- #

def test_map2_not_same_size():
    sum2 = lambda x, y: x + y
    seg11 = Segment([TaggedValue(13, "C")])
    seg21 = Segment([TaggedValue(31, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    seg31 = Segment([TaggedValue(72, "N"), TaggedValue(92, "L"), TaggedValue(42, "L")])
    lt1 = LTree([seg11, seg21, seg31])
    seg22 = Segment([TaggedValue(31, "N"), TaggedValue(47, "L"), TaggedValue(32, "L")])
    lt2 = LTree([seg22])
    pt1 = PTree(lt1)
    pt2 = PTree(lt2)
    with pytest.raises(AssertionError):
        pt1.zip(pt2)


def test_map2():
    sum2 = lambda x, y: x + y
    seg11 = Segment([TaggedValue(1, "C")])
    seg21 = Segment([TaggedValue(1, "N"), TaggedValue(1, "L"), TaggedValue(1, "L")])
    seg31 = Segment([TaggedValue(1, "N"), TaggedValue(1, "L"), TaggedValue(1, "L")])
    lt1 = LTree([seg11, seg21, seg31])
    seg12 = Segment([TaggedValue(2, "C")])
    seg22 = Segment([TaggedValue(2, "N"), TaggedValue(2, "L"), TaggedValue(2, "L")])
    seg32 = Segment([TaggedValue(2, "N"), TaggedValue(2, "L"), TaggedValue(2, "L")])
    lt2 = LTree([seg12, seg22, seg32])
    pt1 = PTree(lt1)
    pt2 = PTree(lt2)
    res = pt1.map2(sum2, pt2).to_seq()
    seg1_exp = Segment([TaggedValue(3, "C")])
    seg2_exp = Segment([TaggedValue(3, "N"), TaggedValue(3, "L"), TaggedValue(3, "L")])
    seg3_exp = Segment([TaggedValue(3, "N"), TaggedValue(3, "L"), TaggedValue(3, "L")])
    exp = LTree([seg1_exp, seg2_exp, seg3_exp]) if pid == 0 else None
    assert res == exp


import pytest

from pyske.core.tree.ltree import LTree, Segment, TaggedValue
from pyske.core.support.errors import IllFormedError


def test_map_empty():
	lt = LTree()
	id_f = lambda x : x
	with pytest.raises(AssertionError):
		lt.map(id_f, id_f)


def test_map_not_empty():
	seg1 = Segment([TaggedValue(13, "C")])
	seg2 = Segment([TaggedValue(31, "N"),TaggedValue(47, "L"),TaggedValue(32, "L")])
	seg3 = Segment([TaggedValue(72, "N"),TaggedValue(92, "L"),TaggedValue(42, "L")])
	lt = LTree([seg1, seg2, seg3])

	plus1 = lambda x : x + 1
	minus1 = lambda x : x - 1
	
	res = lt.map(plus1, minus1)

	seg1_exp = Segment([TaggedValue(12, "C")])
	seg2_exp = Segment([TaggedValue(30, "N"),TaggedValue(48, "L"),TaggedValue(33, "L")])
	seg3_exp = Segment([TaggedValue(71, "N"),TaggedValue(93, "L"),TaggedValue(43, "L")])

	exp = LTree([seg1_exp,seg2_exp,seg3_exp])
	assert res == exp


# -------------------------- #

def test_reduce_empty():
	lt = LTree()
	id_f = lambda x : x
	sum3 = lambda x,y,z : x + y + z
	with pytest.raises(AssertionError):
		lt.reduce(sum3, id_f, sum3, sum3, sum3)


def test_reduce_illformed():
	seg1 = Segment([TaggedValue(13, "C")])
	seg3 = Segment([TaggedValue(72, "N"),TaggedValue(92, "L"),TaggedValue(42, "L")])
	lt = LTree([seg1, seg3])
	id_f = lambda x : x
	sum3 = lambda x,y,z : x + y + z
	with pytest.raises(IllFormedError):
		lt.reduce(sum3, id_f, sum3, sum3, sum3)


def test_reduce():
	seg1 = Segment([TaggedValue(13, "C")])
	seg2 = Segment([TaggedValue(31, "N"),TaggedValue(47, "L"),TaggedValue(32, "L")])
	seg3 = Segment([TaggedValue(72, "N"),TaggedValue(92, "L"),TaggedValue(42, "L")])
	lt = LTree([seg1, seg2, seg3])
	id_f = lambda x : x
	sum3 = lambda x,y,z : x + y + z
	res = lt.reduce(sum3, id_f, sum3, sum3, sum3)
	exp = 13 + 31 + 47 + 32 + 72 + 92 + 42
	assert res == exp


# -------------------------- #

def test_uacc_empty():
	lt = LTree()
	id_f = lambda x : x
	sum3 = lambda x,y,z : x + y + z
	with pytest.raises(AssertionError):
		lt.uacc(sum3, id_f, sum3, sum3, sum3)

def test_uacc_illformed():
	seg1 = Segment([TaggedValue(13, "C")])
	seg3 = Segment([TaggedValue(72, "N"),TaggedValue(92, "L"),TaggedValue(42, "L")])
	lt = LTree([seg1, seg3])
	id_f = lambda x : x
	sum3 = lambda x,y,z : x + y + z
	with pytest.raises(IllFormedError):
		lt.uacc(sum3, id_f, sum3, sum3, sum3)

def test_uacc():
	seg1 = Segment([TaggedValue(13, "C")])
	seg2 = Segment([TaggedValue(31, "N"),TaggedValue(47, "L"),TaggedValue(32, "L")])
	seg3 = Segment([TaggedValue(72, "N"),TaggedValue(92, "L"),TaggedValue(42, "L")])
	lt = LTree([seg1, seg2, seg3])
	id_f = lambda x : x
	sum3 = lambda x,y,z : x + y + z
	res = lt.uacc(sum3, id_f, sum3, sum3, sum3)

	seg1_exp = Segment([TaggedValue(13+31+47+32+72+92+42, "C")])
	seg2_exp = Segment([TaggedValue(31+47+32, "N"),TaggedValue(47, "L"),TaggedValue(32, "L")])
	seg3_exp = Segment([TaggedValue(72+92+42, "N"),TaggedValue(92, "L"),TaggedValue(42, "L")])
	exp = LTree([seg1_exp, seg2_exp, seg3_exp])

	assert exp == res


# -------------------------- #

def test_dacc_empty():
	sum2 = lambda x,y : x + y
	c = 0
	id_f = lambda x : x
	lt = LTree()
	with pytest.raises(AssertionError):
		lt.dacc(sum2, sum2, c, id_f, id_f, sum2, sum2)


def test_dacc():
	sum2 = lambda x,y : x + y
	c = 0
	id_f = lambda x : x
	seg1 = Segment([TaggedValue(13, "C")])
	seg2 = Segment([TaggedValue(31, "N"),TaggedValue(47, "L"),TaggedValue(32, "L")])
	seg3 = Segment([TaggedValue(72, "N"),TaggedValue(92, "L"),TaggedValue(42, "L")])
	lt = LTree([seg1, seg2, seg3])
	res = lt.dacc(sum2, sum2, c, id_f, id_f, sum2, sum2)

	seg1_exp = Segment([TaggedValue(0, "C")]) 
	seg2_exp = Segment([TaggedValue(13, "N"),TaggedValue(13 + 31, "L"),TaggedValue(13 + 31, "L")])
	seg3_exp = Segment([TaggedValue(13, "N"),TaggedValue(13 + 72, "L"),TaggedValue(13 + 72, "L")])
	exp = LTree([seg1_exp, seg2_exp, seg3_exp])

	assert res == exp


# -------------------------- #

def test_zip_not_same_size():
	seg11 = Segment([TaggedValue(13, "C")])
	seg21 = Segment([TaggedValue(31, "N"),TaggedValue(47, "L"),TaggedValue(32, "L")])
	seg31 = Segment([TaggedValue(72, "N"),TaggedValue(92, "L"),TaggedValue(42, "L")])
	lt1 = LTree([seg11, seg21, seg31]) 
	seg12 = Segment([TaggedValue(13, "C")])
	seg22 = Segment([TaggedValue(31, "N"),TaggedValue(47, "L"),TaggedValue(32, "L")])
	lt2 = LTree([seg12, seg22])
	with pytest.raises(AssertionError):
		lt1.zip(lt2)


def test_zip():
	seg11 = Segment([TaggedValue(1, "C")])
	seg21 = Segment([TaggedValue(1, "N"),TaggedValue(1, "L"),TaggedValue(1, "L")])
	seg31 = Segment([TaggedValue(1, "N"),TaggedValue(1, "L"),TaggedValue(1, "L")])
	lt1 = LTree([seg11, seg21, seg31]) 
	seg12 = Segment([TaggedValue(2, "C")])
	seg22 = Segment([TaggedValue(2, "N"),TaggedValue(2, "L"),TaggedValue(2, "L")])
	seg32 = Segment([TaggedValue(2, "N"),TaggedValue(2, "L"),TaggedValue(2, "L")])
	lt2 = LTree([seg12, seg22, seg32])
	res = lt1.zip(lt2)
	seg1_exp = Segment([TaggedValue((1,2), "C")])
	seg2_exp = Segment([TaggedValue((1,2), "N"),TaggedValue((1,2), "L"),TaggedValue((1,2), "L")])
	seg3_exp = Segment([TaggedValue((1,2), "N"),TaggedValue((1,2), "L"),TaggedValue((1,2), "L")])
	exp = LTree([seg1_exp, seg2_exp, seg3_exp])
	assert res == exp


# -------------------------- #

def test_map2_not_same_size():
	sum2 = lambda x,y : x + y
	seg11 = Segment([TaggedValue(13, "C")])
	seg21 = Segment([TaggedValue(31, "N"),TaggedValue(47, "L"),TaggedValue(32, "L")])
	seg31 = Segment([TaggedValue(72, "N"),TaggedValue(92, "L"),TaggedValue(42, "L")])
	lt1 = LTree([seg11, seg21, seg31]) 
	seg12 = Segment([TaggedValue(13, "C")])
	seg22 = Segment([TaggedValue(31, "N"),TaggedValue(47, "L"),TaggedValue(32, "L")])
	lt2 = LTree([seg12, seg22])
	with pytest.raises(AssertionError):
		lt1.map2(sum2, lt2)


def test_map2():
	sum2 = lambda x,y : x + y
	seg11 = Segment([TaggedValue(1, "C")])
	seg21 = Segment([TaggedValue(1, "N"),TaggedValue(1, "L"),TaggedValue(1, "L")])
	seg31 = Segment([TaggedValue(1, "N"),TaggedValue(1, "L"),TaggedValue(1, "L")])
	lt1 = LTree([seg11, seg21, seg31]) 
	seg12 = Segment([TaggedValue(2, "C")])
	seg22 = Segment([TaggedValue(2, "N"),TaggedValue(2, "L"),TaggedValue(2, "L")])
	seg32 = Segment([TaggedValue(2, "N"),TaggedValue(2, "L"),TaggedValue(2, "L")])
	lt2 = LTree([seg12, seg22, seg32])
	res = lt1.map2(sum2, lt2)
	seg1_exp = Segment([TaggedValue(3, "C")])
	seg2_exp = Segment([TaggedValue(3, "N"),TaggedValue(3, "L"),TaggedValue(3, "L")])
	seg3_exp = Segment([TaggedValue(3, "N"),TaggedValue(3, "L"),TaggedValue(3, "L")])
	exp = LTree([seg1_exp, seg2_exp, seg3_exp])
	assert res == exp


import functools


LEFT_LIST = "["
RIGHT_LIST = "]"
SEPARATOR_LIST = ";"


class SList(list):
    """
    An extended definition of lists, including Bird-Meertens Formalism primitives

    ...

    Methods
    -------
    from_str(s, parser):
        Creates a SList from a string
    head()
        Gives the first element of the current instance
    tail()
        Gives the the current instance without its first element
    length()
        Gives the number of element in the current instance
    filter(f)
        Removes all the elements that don't verify a predicate
    empty()
        Indicates if a list is empty
    reverse()
        Reverse a list
    map(f)
        Applies f to every element of the current instance
    reduce(f)
        Reduce the current instance using a reduction function
    scan(f, c)
        Makes an total rightward accumulation of the element on the current instance from an initial value
    scanl(f, c)
        Makes a rightward accumulation of the values from an initial one,
        without considering the last value of the instance
    scanr(f)
        Makes a total leftward accumulation of the values
    scanp(f, c)
         Makes an total lefttward accumulation of the element on the current instance from an initial value
    scanl_last(f, c)
        Makes a rightward accumulation of the values from an initial one,
        considering the last accumulation as an external value
    zip(l)
        Creates a list of pairs from the element of the current instance and another one
    map2(f, l)
        Creates a list of new elements using a function from the element of the current instance and another one
    """

    def __str__(self):
        res = LEFT_LIST
        for i in range(0, self.length()):
            res = res + str(self[i])
            if i != self.length() - 1:
                res = res + SEPARATOR_LIST + " "
        return res + RIGHT_LIST

    @staticmethod
    def from_str(s, parser=int):
        """
        Creates a SList from a string

        Parameters
        ----------
        s : str
            A string representation of the SList
        parser : callable, optional
            A function that transforms a string into a specific type
            By default, string to int
        """
        res = SList([])
        values = s.replace(LEFT_LIST, "").replace(RIGHT_LIST, "").split(SEPARATOR_LIST)
        for v in values:
            res.append(parser(v))
        return res

    def head(self):
        """Gives the first element of the current instance
        """
        if self.empty():
            return None
        else:
            return self[0]

    def tail(self):
        """Gives the the current instance without its first element
        """
        return SList(self[1:])

    def length(self):
        """Gives the number of element in the current instance
        """
        return len(self)

    def filter(self, p):
        """Removes all the elements that don't verify a predicate

        Parameters
        ----------
        p : callable
            A predicate that all elements in the result must verify
        """
        return SList(filter(p, self))

    def empty(self):
        """Indicates if a list is empty
        """
        return self.length() == 0

    def reverse(self):
        """Reverse a list
        """
        rev = SList()
        for i in range(self.length() - 1, -1, -1):
            rev.append(self[i])
        return rev

    def map(self, f):
        """Applies f to every element of the current instance

        Definition:
        map f [x1, x2, ..., xn] = [f(x1), f(x2), ..., f(xn)]

        Parameters
        ----------
        f : callable
            The function to apply to every values of the current instance
        """
        return SList(map(f, self))

    def mapi(self, f):
        """Applies f to every index and element of the current instance

        Definition:
        map f [x0, x1, ..., xn] = [f(0,x0), f(1,x1), ..., f(n,xn)]

        Parameters
        ----------
        f: callable
        The function to apply to every index and element of the current instance
        """
        return SList([f(i, self[i]) for i in range(0, len(self))])

    def reduce(self, f, e=None):
        """Reduce the current instance using a reduction function

        Definition:
        reduce f [x1, x2, ..., xn] e = f(f(f(e, x1), ...), xn)

        Parameters
        ----------
        f : callable
            The used function to reduce the current instance
        e : optional
            Default value for reduction
        """
        if e is None:
            return functools.reduce(f, self)
        else:
            return functools.reduce(f, self, e)

    def scan(self, f, c):
        """Makes total a rightward accumulation of the values from an initial one
        The result of scan is a list of size n+1 where n is the size of self.

        Definition:
            scan f c [x_1, x_2, ..., x_n] = [c, f(c, x_1), f(f(c, x_1), x_2), ..., f(f(...,f(f(c, x_1), x_2)), x_n)]

        Parameters
        ----------
        f : callable
            A function to make a new accumulation from the previous accumulation and a current value
        c
            Initial value for the accumulator
        """
        res = self.copy()
        res.append(c)
        res[0] = c
        for i in range(1, len(res)):
            c = f(c, self[i - 1])
            res[i] = c
        return res

    def scanl(self, f, c):
        """Makes a rightward accumulation of the values from an initial one,
        without considering the last value of the instance
        The result of scanl is a list of size n where n is the size of self.

        Definition:
            scanl f c [] = []
            scanl f c [x_1, x_2, ..., x_n] = [c, f(c, x_1), f(f(c, x_1), x_2), ..., f(f(...,f(f(c, x_1), x_2)), x_n-1)]

        Parameters
        ----------
        f : callable
            A function to make a new accumulation from the previous accumulation and a current value
        c
            Initial value for the accumulator
        """
        res = self.copy()
        for i in range(0, len(res)):
            res[i] = c
            c = f(c, self[i])
        return res

    def scanr(self, f):
        """Makes a rightward accumulation of the values.

        Definition:
            scanr f [x] = [x]
            scanr f [x_1, x_2, ..., x_n] = [x_1, f(x_1, x_2), ..., f(f(f(x_1, x_2), ...), x_n)]

        Parameters
        ----------
        f : callable
            A function to make a new accumulation from the previous accumulation and a current value
        """
        assert (len(self) > 0)
        res = self.copy()
        c = res[0]
        for i in range(1, len(res)):
            c = f(c, self[i])
            res[i] = c
        return res

    def scanl_last(self, f, c):
        """Makes a rightward accumulation of the values from an initial one,
        considering the last accumulation as an external value.
        The result of scanl_last is a list of size n where n is the size of self
        and one additional value corresponding to the total accumulation.

        Definition:
            scanl_last f c [] = ([],c)
            scanl_last f c [x_1, x_2, ..., x_n]
                = [c, f(c, x_1), f(f(c, x_1), x_2), ..., f(f(...,f(f(c, x_1), x_2)), x_n-1)]

        Parameters
        ----------
        f : callable
            A function to make a new accumulation from the previous accumulation and a current value
        c
            Initial value for the accumulator
        """
        res = self.scan(f, c)
        last = res.pop()
        return res, last

    def scanp(self, f, c):
        """Makes a leftward accumulation of the values from an initial one.
        The result of scanp is a list of size n where n is the size of self
        and one additional value corresponding to the total accumulation.

        Definition:
            scanp f c [x_1, x_2, ..., x_n] = [f(x_2, f(x_3, f(..., c))), ..., f(x_n, c), c]

        Parameters
        ----------
        f : callable
            A function to make a new accumulation from the previous accumulation and a current value
            Usually, f is associative.
        c
            Initial value for the accumulator.
            Usually, c is the unit of f, i.e. f(x, c) = f(c, x) = x
        """
        res = self.copy()
        for i in range(len(self), 0, -1):
            res[i - 1] = c
            c = f(self[i - 1], c)
        return res

    def zip(self, l):
        """Creates a list of pairs from the element of the current instance and another sequential list

        Precondition
        -------------
        The lengths of self and l should be equal.

        Parameters
        ----------
        l : list
            A list to merge the values of the current instance with
        """
        assert (len(self) == len(l))
        return SList([(x, y) for (x, y) in zip(self, l)])

    def map2(self, f, l):
        """Creates a list of new elements using a function applied to the elements of the current
        instance and another sequential list

        Precondition
        -------------
        The lengths of self and l should be equal.

        Parameters
        ----------
        f : callable
             A function to zip values
        l : list
            The second list to zip with the current instance
        """
        assert (len(self) == len(l))
        return SList([f(x, y) for (x, y) in zip(self, l)])



import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
     name='pyske',  
     version='0.1.1',
     scripts=['psk'] ,
     license='MIT',
     author="Jolan Philippe and Frederic Loulergue",
     author_email="jp2589@nau.edu, frederic.loulergue@nau.edu",
     description="A package for Python Skeletons using mpi4py",
     long_description=long_description,
   long_description_content_type="text/markdown",
     packages=  setuptools.find_packages(),
     install_requires=['mpi4py'],
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )



from pyske.core.list.slist import SList
from pyske.core.list.plist import PList
from pyske.core.support.parallel import *


def app(l1, l2):
    return SList(l1 + l2)


msg = "hello world!"
pl1 = PList.init(lambda i: msg[i], len(msg))
pl2 = PList.init(lambda x: x, len(msg))
pl3 = pl1.map(lambda x: x.capitalize())
pl4 = pl3.zip(pl2)
pl5 = pl4.map(lambda x: x[0])
pl6 = pl5.mapi(lambda i, x: (i, x))
pl7 = pl6.map(lambda x: x[1])
pl8 = pl7.map(lambda x: 1)
n = pl8.reduce(lambda x, y: x + y, 0)
pl9 = pl7.get_partition()
pl10 = pl9.map(lambda l: SList(l).filter(lambda c: c != 'O')).flatten()
pl11 = PList.from_seq(["Hello World!"])

filtered = pl10.get_partition().reduce(app, []).reduce(lambda x, y: x + y)
s1 = pl9.reduce(app).reduce(lambda x, y: x + y)
s2 = pl11.to_seq()[0]

at_root(lambda:
        print(f'Capitalized: \t{s1}\n'
              f'Identity:\t{s2}\n'
              f'Length:\t\t{n}\n'
              f'Filtered:\t{filtered}'))



def prefix(tree):
	# ---------- #

	f = lambda a : (0,1)
	id_f = lambda x: x

	# ---------- #

	phi = lambda b: (1, 0, 0, 1)

	def k(l, b, r):
		(ll, ls) = l
		(rl, rs) = r
		return (ls, ls + 1 + rs)
		

	def psi_n(l, b, r):
		(ll, ls) = l
		(b0, b1, b2, b3) = b
		(rl, rs) = r
		res_1 = b0 * ls + b1 * (ls + rs + 1) + b2
		res_2 = ls + 1 + rs + b3
		return (res_1, res_2)


	def psi_l(l, b, r):
		(l0, l1, l2, l3) = l
		(b0, b1, b2, b3) = b
		(rl, rs) = r
		res_0 = 0
		res_1 = b0 + b1
		res_2 = (b0 + b1) * l3 + b1 * (1 + rs) + b2
		res_3 = l3 + 1 + rs + b3
		return (res_0, res_1, res_2, res_3)


	def psi_r(l, b, r):
		(ll, ls) = l
		(b0, b1, b2, b3) = b
		(r0, r1, r2, r3) = r
		res_0 = 0
		res_1 = b1
		res_2 = b1 * r3 + b0 * ls + b1 * (1 + ls) + b2
		res_3 = r3 + 1 + ls + b3
		return (res_0, res_1, res_2, res_3)

	# ---------- #

	def gl(c, b):
		(bl, bs) = b
		return c + 1

	def gr(c, b):
		(bl, bs) = b
		return c + bl + 1

	def phi_l(b):
		(bl, bs) = b
		return 1

	def phi_r(b):
		(bl, bs) = b
		return bl+1

	sum2 = lambda x,y : x + y
	# ---------- #

	mapped = tree.map(f,id_f)
	tree2 = mapped.uacc(k, phi, psi_n, psi_l, psi_r)
	return tree2.dacc(gl, gr, 0, phi_l, phi_r, sum2, sum2)